from vc.client.schemas_base import ResponseData


class ClientException(Exception):
    def __init__(self, data, status_code, url):
        self.status_code = status_code
        self.data = data
        self.url = url

    def __str__(self):
        return f"{self.status_code}: {self.data} {self.url}"

    def to_dict(self):
        return dict(loc=["api"], msg=self.data, type=self.status_code)


def response_builder(expected_code=200, response_schema=ResponseData, data_schema=None):
    def wrapper(func):
        def builder(*args, **kwargs):
            data, status_code, url = func(*args, **kwargs)
            response = response_schema(status_code=status_code, url=url, **data)
            if status_code == expected_code:
                if data_schema:
                    try:
                        if isinstance(data, list):
                            response.data = data
                        else:
                            response.data = data_schema(**data)
                    except Exception as e:
                        print(data)
                        raise e
            else:
                raise ClientException(data, status_code, url)
            return response

        return builder

    return wrapper


from datetime import datetime
from decimal import Decimal
from typing import Optional, Generic, TypeVar, List, Dict
from pydantic import BaseModel, Extra, root_validator
from pydantic.generics import GenericModel
from .utils import decrypt

T = TypeVar('T')


class ResponseInfo(BaseModel):

    class Config:
        extra = Extra.allow


class UserBase(ResponseInfo):
    name: Optional[str]
    email: Optional[str]
    status: Optional[str]
    surname: Optional[str]
    job_title: Optional[str]
    custom_reference_id: Optional[str]


class UserResponse(UserBase):
    id: Optional[str]


class Order(GenericModel, Generic[T]):
# {
# "id":"6a5e0887-eaf6-4f49-95cb-6b7f245d04f3",
# "status":"PLACED","creation_time":"2021-06-29T10:42:06",
# "last_update_time":"2021-06-29T10:42:06","is_valid":true,"total_paid_amount":0.000000,
# "total_paid_currency":"EUR",
# "items":[{"id":"87d95150-41f3-46a5-9501-ca1b368696ca","itemType":"WALLET","category":"WALLET"}]}
    id: str
    status: str
    creation_time: datetime
    last_update_time: datetime
    is_valid: bool
    total_paid_amount: Decimal
    total_paid_currency: str
    items: Optional[List[T]]


class OrderItem(ResponseInfo):
    id: Optional[str]
    itemType: Optional[str]
    category: Optional[str]


class CardResponse(ResponseInfo):
    id: str
    label: Optional[str]
    name: Optional[str]
    status: str
    sensitive_data: Optional[Dict]
    expiration_date: datetime
    creation_time: datetime
    wallet_id: Optional[str]
    type: Optional[str]
    owner_type: str
    owner_public_id: Optional[str]
    masked_pan: str
    pan: Optional[str] = None
    cvv: Optional[str] = None

    @root_validator(pre=True)
    def validate_root(cls, values):
        if values.get("sensitive_data"):
            if values['sensitive_data'].get("encrypted_full_pan"):
                values["pan"] = decrypt(values['sensitive_data'].get("encrypted_full_pan"))
            if values['sensitive_data'].get("encrypted_cvv"):
                values["cvv"] = decrypt(values['sensitive_data'].get("encrypted_cvv"))
        if not values.get("pan"):
            values["pan"] = values.get("masked_pan")
        return values


class PaginateList(GenericModel, Generic[T]):
    results: List[T]
    current_page: int
    page_size: int
    results_size: int
    total: int


class WalletBase(ResponseInfo):
    id: Optional[str]
    name: Optional[str]
    primary_user_public_id: Optional[str]
    available_amount: Optional[Decimal]
    blocked_amount: Optional[Decimal]


import base64
import logging
import time

from cryptography.exceptions import AlreadyFinalized
from cryptography.exceptions import UnsupportedAlgorithm
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import padding


logger = logging.getLogger(__name__)


def decrypt(ciphertext: str, ):
    from vc.manager import Soldo
    try:
        # load private key
        with open(Soldo.settings.PATH_RSA_PRIVATE, "rb") as key_file:
            privateKey = serialization.load_pem_private_key(
                key_file.read(),
                password=None,
                backend=default_backend()
            )

        # decrypt string
        plaintext = privateKey.decrypt(
            base64.b64decode(ciphertext),
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None
            )
        )

        return plaintext.decode()
    except Exception as e:
        logger.exception(f"Decode failed: {e}")


# function to be used to generate the hashed fingerprint, input argument is the fingerprint order concatenated string
def fingerprintHash(fingerprint):
    try:
        # get digest instance
        digest = hashes.Hash(
            algorithm=hashes.SHA512(),
            backend=default_backend()
        )

        # create hash
        digest.update(fingerprint.encode('utf-8'))
        hash_bytes = digest.finalize()

        # convert/encode in hex
        return hash_bytes.hex()

    except (UnsupportedAlgorithm, AlreadyFinalized):
        logger.exception("Hashing failed")


# function to be used to generate the fingerprint signature, input arguments are the hashed fingerprint to be signed and the private 2048 RSA key file path
def fingerprintSignature(fingerprintH):
    from vc.manager import Soldo
    try:
        # load private key
        with open(Soldo.settings.PATH_RSA_PRIVATE, "rb") as key_file:
            privateKey = serialization.load_pem_private_key(
                key_file.read(),
                password=None,
                backend=default_backend()
            )

        # sign string
        signature = privateKey.sign(
            data=fingerprintH.encode('utf-8'),
            padding=padding.PKCS1v15(),
            algorithm=hashes.SHA512()
        )

        return base64.b64encode(signature)
    except UnsupportedAlgorithm:
        logger.exception("Signing failed")


def request_timestamp():
    return int(time.time())
